(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Format

open Readable_types
open Readable_annotations
open Asmprogram
open Util

let label_new_number = ref 0

let get_new_label () =
	let v = !label_new_number in
	label_new_number := !label_new_number + 1;
	let s = string_of_int v in
	".L_____TAL_ANNOTS_____" ^ s

let init_label init_num =
	label_new_number := init_num

let convert_label ppf label =
	fprintf ppf "%s:\n" label

let output_reg () reg =
	if reg = "" then
		""
	else
		sprintf "%%%s" reg

let output_reg_comma () reg =
	if reg = "" then
		""
	else
		sprintf ", %%%s" reg

let output_str_comma () s =
	if s = "" then
		""
	else
		sprintf ", %s" s

let output_op ppf op =
	match op with
	| OpImd (s, _) ->
		fprintf ppf "$%s" s
	| OpReg s ->
		fprintf ppf "%%%s" s
	| OpMem (s, _) ->
		fprintf ppf "%s" s

let output_jump_op ppf op =
	match op with
	| OpImd (s, _) ->
		fprintf ppf "%s" s
	| OpReg s ->
		fprintf ppf "%%%s" s
	| OpMem (s, _) ->
		fprintf ppf "%s" s

let get_mem_annot_op op =
	match op with
	| OpMem (_, mem_annot) ->
		Some mem_annot
	| _ ->
		None

let get_mem_annot rinst =
	match rinst with
	| Inst0 _ ->
		None
	| Inst1 (_, op) ->
		get_mem_annot_op op
	| Inst2 (_, op1, op2) ->
	begin
		let r = get_mem_annot_op op1 in
		match r with
		| None ->
			get_mem_annot_op op2
		| _ ->
			r
	end
	| Inst3 (_, op1, op2, op3) ->
	begin
		let r = get_mem_annot_op op1 in
		match r with
		| None ->
		begin
			let r = get_mem_annot_op op2 in
			match r with
			| None ->
				get_mem_annot_op op3
			| _ ->
				r
		end
		| _ ->
			r
	end

let get_sign_annot_op op =
	match op with
	| OpImd (_, sign) ->
		sign
	| _ ->
		false

let get_sign_annot rinst =
	match rinst with
	| Inst0 _ ->
		false
	| Inst1 (_, op) ->
		get_sign_annot_op op
	| Inst2 (_, op1, op2) ->
		let r = get_sign_annot_op op1 in
		r || get_sign_annot_op op2
	| Inst3 (_, op1, op2, op3) ->
		let r1 = get_sign_annot_op op1 in
		let r2 = get_sign_annot_op op2 in
		let r3 = get_sign_annot_op op3 in
		r1 || r2 || r3

let convert_list ppf conv lst =
	let n = List.length lst in
	fprintf ppf ".uleb128 %d\n" n;
	List.iter (conv ppf) lst

let convert_nativeint ppf ni =
	fprintf ppf ".long 0x%nx\n" ni

let convert_string ppf name =
	fprintf ppf ".uleb128 %d\n" (String.length name);
	fprintf ppf ".ascii \"%s\"\n" name

let convert_bool ppf b =
	if b then
		fprintf ppf ".byte 'T'\n"
	else
		fprintf ppf ".byte 'F'\n"

let convert_tv ppf tv =
	let TV(name) = tv in
	convert_string ppf name

let convert_tvs ppf tvs =
	convert_list ppf convert_tv tvs

let rec convert_integer ppf inte =
	match inte with
	| IConst s ->
(*
		fprintf ppf ".byte 'C'\n";
		fprintf ppf ".long %s\n" s
*)
		let sign = Util.get_sign s in
		if sign then
			fprintf ppf ".byte 'S'\n"
		else
			fprintf ppf ".byte 'U'\n"
		;
		fprintf ppf ".long %s\n" s
	| IVar tv ->
		fprintf ppf ".byte 'V'\n";
		convert_tv ppf tv

let convert_arith_op ppf aop =
	match aop with
	| A_ADD ->
		fprintf ppf ".byte '+'\n"
	| A_MUL ->
		fprintf ppf ".byte '*'\n"

let convert_cmp_op_raw ppf = function
	| C_EQ ->
		fprintf ppf ".byte '='\n"
	| C_NEQ ->
		fprintf ppf ".byte 'N'\n"
	| C_LT ->
		fprintf ppf ".byte '<'\n"
	| C_GT ->
		fprintf ppf ".byte '>'\n"
	| C_LE ->
		fprintf ppf ".byte 'L'\n"
	| C_GE ->
		fprintf ppf ".byte 'G'\n"

let convert_cmp_op ppf cop =
	let (cop_r, signed) = cop in
	convert_cmp_op_raw ppf cop_r;
	convert_bool ppf signed

let rec convert_integer_expr ppf i_e =
	match i_e with
	| IEInte inte ->
		fprintf ppf ".byte 'I'\n";
		convert_integer ppf inte
	| IEArith (aop, i_e1, i_e2) ->
		fprintf ppf ".byte 'A'\n";
		convert_arith_op ppf aop;
		convert_integer_expr ppf i_e1;
		convert_integer_expr ppf i_e2

let convert_integer_cstrt ppf i_c =
	match i_c with
	| ICCmp (cop, i_e1, i_e2) ->
		fprintf ppf ".byte 'C'\n";
		convert_cmp_op ppf cop;
		convert_integer_expr ppf i_e1;
		convert_integer_expr ppf i_e2

let convert_integer_cstrts ppf ics =
	convert_list ppf convert_integer_cstrt ics

let begin_annots ppf =
	let label = get_new_label () in
	fprintf ppf "%s\n" (label ^ ":");
	fprintf ppf ".section .annots\n";
	fprintf ppf ".byte 'A'\n";
	fprintf ppf "%s\n" (".long " ^ label)

let end_annots ppf =
	fprintf ppf ".previous\n"

(*
let convert_annot_mem ppf base scale index offset =
*)
let convert_annot_mem ppf base offset =
	begin_annots ppf;
	fprintf ppf ".byte 'M'\n";
	convert_integer ppf base;
	convert_nativeint ppf offset;
	end_annots ppf

let convert_mem_annot ppf mem_annot =
(*
	let (base, scale, index, offset) = mem_annot in
*)
	let (base, offset) = mem_annot in
	convert_annot_mem ppf base offset

let convert_mem_annot_opt ppf mem_annot_opt =
	match mem_annot_opt with
	| None ->
		()
	| Some mem_annot ->
		convert_mem_annot ppf mem_annot

let convert_sign_annot ppf =
	begin_annots ppf;
	fprintf ppf ".byte '-'\n";
	end_annots ppf

let convert_sign_annot ppf sign =
	if sign then
		convert_sign_annot ppf

let is_jump name =
	name = "jmp"
	|| name = "call"
	|| name = "jb"
	|| name = "jbe"
	|| name = "ja"
	|| name = "jae"
	|| name = "jz"
	|| name = "jnz"

let convert_rinst ppf rinst =
	let mem_annot_opt = get_mem_annot rinst in
	convert_mem_annot_opt ppf mem_annot_opt;

	let sign_annot = get_sign_annot rinst in
	convert_sign_annot ppf sign_annot;

	match rinst with
	| Inst0 name ->
		fprintf ppf "\t%s\n" name
	| Inst1 (name, op) ->
		if is_jump name then
			fprintf ppf "\t%s %a\n" name output_jump_op op
		else
			fprintf ppf "\t%s %a\n" name output_op op
	| Inst2 (name, op1, op2) ->
		fprintf ppf "\t%s %a, %a\n" name output_op op1 output_op op2
	| Inst3 (name, op1, op2, op3) ->
		fprintf ppf "\t%s %a, %a, %a\n" name output_op op1 output_op op2 output_op op3

let convert_kind ppf k =
	match k with
	| KInt ->
		fprintf ppf ".byte 'I'\n"
	| KStack ->
		fprintf ppf ".byte 'K'\n"
	| KStore ->
		fprintf ppf ".byte 'S'\n"

let convert_tv_and_kind ppf p =
	let (tv, kind) = p in
	convert_tv ppf tv;
	convert_kind ppf kind

let convert_tvs_and_kinds ppf tvs =
	convert_list ppf convert_tv_and_kind tvs

let convert_register_kind ppf reg_kind =
	match reg_kind with
	| RK_L ->
		fprintf ppf ".byte 'L'\n"
	| RK_H ->
		fprintf ppf ".byte 'H'\n"
	| RK_X ->
		fprintf ppf ".byte 'X'\n"
	| RK_E ->
		fprintf ppf ".byte 'E'\n"

let convert_register ppf (reg_kind, inte) =
	convert_register_kind ppf reg_kind;
	convert_integer ppf inte

let convert_operand_size ppf = function
	| Instructions.OpSize8 ->
		fprintf ppf ".byte '1'\n"
	| Instructions.OpSize16 ->
		fprintf ppf ".byte '2'\n"
	| Instructions.OpSize32 ->
		fprintf ppf ".byte '4'\n"
	| Instructions.OpSize48 ->
		fprintf ppf ".byte '6'\n"
	| Instructions.OpSize64 ->
		fprintf ppf ".byte '8'\n"
	| Instructions.OpSize128 ->
		fprintf ppf ".byte '0'\n"

let convert_flags ppf flags =
	convert_integer ppf flags.f_rst;
	convert_integer ppf flags.f_op1;
	convert_integer ppf flags.f_op2;
	convert_operand_size ppf flags.f_size

let convert_flags_opt ppf = function
	| None ->
		fprintf ppf ".byte 'N'\n"
	| Some flags ->
	begin
		fprintf ppf ".byte 'S'\n";
		convert_flags ppf flags
	end
		


let convert_registers ppf regs =
	convert_register ppf regs.eax;
	convert_register ppf regs.ecx;
	convert_register ppf regs.edx;
	convert_register ppf regs.ebx;
	convert_register ppf regs.esp;
	convert_register ppf regs.ebp;
	convert_register ppf regs.esi;
	convert_register ppf regs.edi;
	convert_flags_opt ppf regs.flags_opt

let convert_tuple_elem_type ppf (sz, inte) =
	convert_operand_size ppf sz;
	convert_integer ppf inte

let convert_tuple_type ppf tt =
	convert_list ppf convert_tuple_elem_type tt

let rec convert_stack_type ppf st =
	match st with
	| StkNull ->
		fprintf ppf ".byte 'N'\n"
	| StkVar tv ->
		fprintf ppf ".byte 'V'\n";
		convert_tv ppf tv
	| StkCons (inte, st) ->
		fprintf ppf ".byte 'C'\n";
		convert_integer ppf inte;
		convert_stack_type ppf st

let convert_rich_stack_type ppf stk =
	let freelen = stk.rstk_freelen in
	let stk = stk.rstk_stk in
	convert_integer ppf freelen;
	convert_stack_type ppf stk

let rec convert_atom ppf a =
	let (inte, ht) = a in
	convert_integer ppf inte;
	convert_heap_type ppf ht

and convert_code_atom ppf a =
	let (inte, st) = a in
	convert_integer ppf inte;
	convert_code_type ppf st

and convert_store_elem ppf se =
	match se with	
	| SEAtom a ->
		fprintf ppf ".byte 'A'\n";
		convert_atom ppf a
	| SEVar tv ->
		fprintf ppf ".byte 'V'\n";
		convert_tv ppf tv

and convert_store ppf store =
	convert_list ppf convert_store_elem store

and convert_code_store ppf c_str =
	convert_list ppf convert_code_atom c_str

and convert_cstrt ppf cstrt =
	convert_integer_cstrts ppf cstrt.c_ics;
	convert_store ppf cstrt.c_store;
	convert_code_store ppf cstrt.c_code_store

and convert_prefix ppf prefix =
	convert_tvs_and_kinds ppf prefix.p_tvs;
	convert_list ppf convert_cstrt prefix.p_cstrts

and convert_state_type ppf st =
	convert_prefix ppf st.st_prefix;
	convert_registers ppf st.st_registers

and convert_pack_type ppf pt =
	convert_prefix ppf pt.pt_prefix;
	convert_tuple_type ppf pt.pt_tuple

and convert_concrete_type ppf ct =
	match ct with
	| Pack pt ->
		fprintf ppf ".byte 'P'\n";
		convert_pack_type ppf pt
	| Tuple tt ->
		fprintf ppf ".byte 'T'\n";
		convert_tuple_type ppf tt

and convert_type ppf typ =
	match typ with
	| Name nt ->
		fprintf ppf ".byte 'N'\n";
		convert_named_type ppf nt
	| Concrete ct ->
		fprintf ppf ".byte 'C'\n";
		convert_concrete_type ppf ct

and convert_array_type ppf at =
	let t = at.at_type in
	let len = at.at_len in
	convert_type ppf t;
	convert_integer ppf len

and convert_heap_type ppf ht =
	match ht with
	| Array at ->
		fprintf ppf ".byte 'A'\n";
		convert_array_type ppf at
	| Stack stk ->
		fprintf ppf ".byte 'S'\n";
		convert_rich_stack_type ppf stk

and convert_code_type ppf ct =
	match ct with
	| NameCode nt ->
		fprintf ppf ".byte 'N'\n";
		convert_named_type ppf nt
	| State st ->
		fprintf ppf ".byte 'S'\n";
		convert_state_type ppf st

and convert_named_type ppf nt =
	let (name, cs) = nt in
	convert_string ppf name;
	convert_constructors ppf cs

and convert_constructor ppf c =
	match c with
	| CInt it ->
		fprintf ppf ".byte 'I'\n";
		convert_integer ppf it
	| CStack st ->
		fprintf ppf ".byte 'K'\n";
		convert_stack_type ppf st
	| CStore st ->
		fprintf ppf ".byte 'S'\n";
		convert_store ppf st

and convert_constructors ppf cs =
	convert_list ppf convert_constructor cs

let convert_named_type_body ppf bdy =
	match bdy with
	| NtdConcrete ct ->
		fprintf ppf ".byte 'C'\n";
		convert_concrete_type ppf ct
	| NtdState st ->
		fprintf ppf ".byte 'S'\n";
		convert_state_type ppf st

let convert_named_type_definition ppf ntd =
	convert_tvs_and_kinds ppf ntd.ntd_tvs;
	convert_named_type_body ppf ntd.ntd_body

let convert_ainst ppf ainst =
	match ainst with
	| A_ROLL (inte, nt) ->
		fprintf ppf ".byte 'R'\n";
		convert_integer ppf inte;
		convert_named_type ppf nt
	| A_UNROLL inte ->
		fprintf ppf ".byte 'U'\n";
		convert_integer ppf inte
	| A_PACK (inte, pt, cs) ->
		fprintf ppf ".byte 'P'\n";
		convert_integer ppf inte;
		convert_pack_type ppf pt;
		convert_constructors ppf cs
	| A_UNPACK (dbg_info, inte) ->
		fprintf ppf ".byte 'O'\n";
		convert_integer ppf inte;
		convert_list ppf convert_string dbg_info
	| A_SPLIT (dbg_info, inte1, inte2) ->
		fprintf ppf ".byte 'S'\n";
		convert_integer ppf inte1;
		convert_integer ppf inte2;
		convert_list ppf convert_string dbg_info
	| A_CONCAT (dbg_info, inte1, inte2) ->
		fprintf ppf ".byte 'C'\n";
		convert_integer ppf inte1;
		convert_integer ppf inte2;
		convert_list ppf convert_string dbg_info
	| A_TUPLE_SPLIT (dbg_info, inte1, inte2) ->
		fprintf ppf ".byte 'L'\n";
		convert_integer ppf inte1;
		convert_integer ppf inte2;
		convert_list ppf convert_string dbg_info
	| A_TUPLE_CONCAT (inte1, inte2) ->
		fprintf ppf ".byte 'N'\n";
		convert_integer ppf inte1;
		convert_integer ppf inte2
	| A_CAST_TO_TUPLE (dbg_info, inte) ->
		fprintf ppf ".byte 'E'\n";
		convert_integer ppf inte;
		convert_list ppf convert_string dbg_info
	| A_CAST_TO_ARRAY (dbg_info, inte) ->
		fprintf ppf ".byte 'Y'\n";
		convert_integer ppf inte;
		convert_list ppf convert_string dbg_info
	| A_CAST_TO_STACK (dbg_info, inte) ->
		fprintf ppf ".byte 'K'\n";
		convert_integer ppf inte;
		convert_list ppf convert_string dbg_info

let convert_annot_dbg_newnames ppf dbg_names =
	begin_annots ppf;
	fprintf ppf ".byte 'D'\n";
	convert_list ppf convert_string dbg_names;
	end_annots ppf

let convert_inst ppf inst =
	match inst with
	| IRaw (rinst, dbg_names) ->
		if dbg_names <> [] then convert_annot_dbg_newnames ppf dbg_names;
		convert_rinst ppf rinst
	| ILabel label ->
		convert_label ppf label
	| IAnnot ainst ->
		begin_annots ppf;
		convert_ainst ppf ainst;
		end_annots ppf

let convert_insts ppf insts =
	List.iter (convert_inst ppf) insts


let get_begin_label blk_name =
	if blk_name = "" then ""
	else
		blk_name ^ ":"

let get_end_label blk_name =
	if blk_name = "" then ""
	else
		".L" ^ blk_name ^ "_____TAL_ASM_____END:"


let convert_label_common converter section_name ppf name a =
	fprintf ppf ".section %s\n" section_name;
	fprintf ppf ".long %s\n" name;
	fprintf ppf ".long .L%s_____TAL_ASM_____END - %s\n" name name ;
	converter ppf a;
	fprintf ppf ".previous\n\n"

let convert_data_label = convert_label_common convert_heap_type ".types"
let convert_code_label = convert_label_common convert_state_type ".codetypes"

let convert_label ppf lab =
	let name = lab.l_name in
	let lt = lab.l_type in

	match lt with
	| LTData rt ->
		convert_data_label ppf name rt
	| LTCode st ->
		convert_code_label ppf name st

let convert_block ppf block =
	match block with
	| ALabelType lab ->
		convert_label ppf lab
	| ANamedType nt ->
		fprintf ppf ".section .namedtypes\n";
		convert_string ppf nt.nt_name;
		convert_named_type_definition ppf nt.nt_definition;
		fprintf ppf ".previous\n\n"
	| AProgram (blk_name, insts) ->
		fprintf ppf ".section .annots\n";
		fprintf ppf ".long %s\n" (if blk_name = "" then "0" else blk_name);
		fprintf ppf ".previous\n";
	 	fprintf ppf "%s\n" (get_begin_label blk_name);
		convert_insts ppf insts;
		fprintf ppf "%s\n" (get_end_label blk_name);
		fprintf ppf ".section .annots\n";
		fprintf ppf ".byte 'E'\n";
		fprintf ppf ".previous\n"

let convert ppf blocks =
	init_label 0;
	List.iter (convert_block ppf) blocks
