(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Instructions
open Readable_pseudo_insts
open Readable_annotations


open Typing_error

open Util


let get_addresses annots =
	List.fold_left (fun lst annot ->
		match annot with
		| A_ADDRESS ad ->
			ad :: lst
		| _ ->
			lst
	) [] annots

let get_dbg_names annots = 
	List.fold_right (fun annot lst ->
		match annot with
		| A_NEWNAMES names ->
			names @ lst
		| _ ->
			lst
	) annots []

let get_signed annots = 
	List.exists (fun annot ->
		annot = A_SIGNED
	) annots

let op_to_p_op op va annots =
	match op with
	| Op_Imd (imd, sz) ->
		let signed = get_signed annots in
		let imd = if signed then
				Big_int_util.big_int_of_nativeint_signed imd
			else
				Big_int_util.big_int_of_nativeint_unsigned imd
		in
		P_Op_Imd (imd, sz)
	| Op_Reg reg ->
		P_Op_Reg reg
	| Op_Mem mop ->
		let ads = get_addresses annots in
		match ads with
		| [ad] ->
			P_Op_Mem (mop, ad)
		| [] ->
			failwithf_type_error "0x%08nx: memory operand must be annotated." va
		| _ ->
			failwithf_type_error "0x%08nx: memory operand was annotated with two or more annotations." va

let annot_inst_to_pinsts va annots annot_inst rst =
	let pinst = match annot_inst with
	| A_ROLL (inte, nt) ->
		P_ROLL (inte, nt) 
	| A_UNROLL inte ->
		P_UNROLL inte
	| A_PACK (inte, pt, cs) ->
		P_PACK (inte, pt, cs)
	| A_UNPACK (dbg_info, inte) ->
		P_UNPACK (dbg_info, inte)
	| A_SPLIT (dbg_info, inte1, inte2) ->
		P_SPLIT (dbg_info, inte1, inte2)
	| A_CONCAT (dbg_info, inte1, inte2) ->
		P_CONCAT (dbg_info, inte1, inte2)
	| A_TUPLE_SPLIT (dbg_info, inte1, inte2) ->
		P_TUPLE_SPLIT (dbg_info, inte1, inte2)
	| A_TUPLE_CONCAT (inte1, inte2) ->
		P_TUPLE_CONCAT (inte1, inte2)
	| A_CAST_TO_TUPLE (dbg_info, inte) ->
		P_CAST_TO_TUPLE (dbg_info, inte)
	| A_CAST_TO_ARRAY (dbg_info, inte) ->
		P_CAST_TO_ARRAY (dbg_info, inte)
	| A_CAST_TO_STACK (dbg_info, inte) ->
		P_CAST_TO_STACK (dbg_info, inte)
	in
		(* The size of annotation-instructions is 0 *)
		(va, va, pinst) :: rst

let annot_insts_to_pinsts va annots annot_insts rst =
	List.fold_right (annot_inst_to_pinsts va annots) annot_insts rst

let convert_POP op va annots =
	let p_op = op_to_p_op op va annots in
	let dbg_names = get_dbg_names annots in
	P_POP (dbg_names, p_op)

let convert_POPF va annots =
	let dbg_names = get_dbg_names annots in
	P_POPF dbg_names

let convert_PUSH op va annots =
	let p_op = op_to_p_op op va annots in
	let dbg_names = get_dbg_names annots in
	P_PUSH (dbg_names, p_op)

let convert_PUSHF va annots =
	let dbg_names = get_dbg_names annots in
	P_PUSHF dbg_names

let convert_MOV op1 op2 va annots =
	let p_op1 = op_to_p_op op1 va annots in
	let p_op2 = op_to_p_op op2 va annots in
	P_MOV (p_op1, p_op2)

let convert_ADD op1 op2 va annots =
	let p_op1 = op_to_p_op op1 va annots in
	let p_op2 = op_to_p_op op2 va annots in
	let dbg_names = get_dbg_names annots in
	P_ADD (dbg_names, p_op1, p_op2)

let convert_SUB op1 op2 va annots =
	let p_op1 = op_to_p_op op1 va annots in
	let p_op2 = op_to_p_op op2 va annots in
	let dbg_names = get_dbg_names annots in
	P_SUB (dbg_names, p_op1, p_op2)

let convert_CMP op1 op2 va annots =
	let p_op1 = op_to_p_op op1 va annots in
	let p_op2 = op_to_p_op op2 va annots in
	let dbg_names = get_dbg_names annots in
	P_CMP (dbg_names, p_op1, p_op2)

let convert_DEC op va annots =
	let p_op = op_to_p_op op va annots in
	let dbg_names = get_dbg_names annots in
	P_DEC (dbg_names, p_op)

let convert_IMUL op1 op2 op3 va annots =
	let p_op1 = op_to_p_op op1 va annots in
	let p_op2 = op_to_p_op op2 va annots in
	let p_op3 = op_to_p_op op3 va annots in
	let dbg_names = get_dbg_names annots in
	P_IMUL (dbg_names, p_op1, p_op2, p_op3)

let convert_Jcc cc op va annots =
	let p_op = op_to_p_op op va annots in
	P_Jcc (cc, p_op)

let convert_CALLN op va annots =
	let p_op = op_to_p_op op va annots in
	P_CALLN p_op

let convert_JMPN op va annots =
	let p_op = op_to_p_op op va annots in
	P_JMPN p_op

let opcode_to_pinst opcode va annots =
	match opcode with
	| Inst_POP op ->
		convert_POP op va annots
	| Inst_POPF ->
		convert_POPF va annots
	| Inst_PUSH op ->
		convert_PUSH op va annots
	| Inst_PUSHF ->
		convert_PUSHF va annots
	| Inst_MOV (op1, op2) ->
		convert_MOV op1 op2 va annots
	| Inst_ADD (op1, op2) ->
		convert_ADD op1 op2 va annots
	| Inst_SUB (op1, op2) ->
		convert_SUB op1 op2 va annots
	| Inst_CMP (op1, op2) ->
		convert_CMP op1 op2 va annots
	| Inst_DEC op ->
		convert_DEC op va annots
	| Inst_IMUL (op1, op2, op3) ->
		convert_IMUL op1 op2 op3 va annots
	| Inst_Jcc (cc, op) ->
		convert_Jcc cc op va annots
	| Inst_RET ->
		P_RET
	| Inst_CALLN op ->
		convert_CALLN op va annots
	| Inst_JMPN op ->
		convert_JMPN op va annots
	| _ ->
		failwithf_type_error "0x%08nx: Unsupported instruction." va

let add_fall va annots pinsts =
	if List.mem A_FALL annots then
		pinsts @ [(va, va, P_FALL)]
	else
		pinsts

let iinfo_to_pinsts va annots_map rst iinfo =
	let off = Nativeint.of_int iinfo.inst_off in
	let sz = Nativeint.of_int iinfo.inst_size in
	let opcode = iinfo.inst_opcode in
	let va = Nativeint.add va off in
	let nva = Nativeint.add va sz in

	let annot_insts, annots =
	try
		let all_annots = NativeintMap.find va annots_map in
		all_annots.annot_insts, all_annots.annot_attrs
	with Not_found ->
		[], [] in

	let pinst = opcode_to_pinst opcode va annots in
	let pinsts = [(va, nva, pinst)] in
	let pinsts = add_fall nva annots pinsts in
	let pinsts = annot_insts_to_pinsts va annots annot_insts pinsts in

	rst @ pinsts

let iinfos_to_pinsts va iinfos annots_map =
	List.fold_left (iinfo_to_pinsts va annots_map) [] iinfos
