(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Types

(**************************************
	Convert routines
	from De Bruijin's notation Type
	to Human-Readable Type
 **************************************)

let find_tv tvs i =
	let len = List.length tvs in
	if (0 <= i && i < len) then
		List.nth tvs i
	else
		Readable_types.TV("??_" ^ (string_of_int i))

let conv_type_variable tvs tv =
	let TV(i) = tv in
	find_tv tvs i

let rec conv_integer tvs = function
	| IConst i ->
(*
		let s = Format.sprintf "0x%nx" i in
*)
		let sign = Big_int.sign_big_int i < 0 in
		let i = Big_int.abs_big_int i in

(*
		let s = Format.sprintf "0x%s" (Big_int_util.string_of_big_int_x i) in
*)
		let s = Format.sprintf "%s" (Big_int_util.string_of_big_int_x i) in
		let s = if sign then
				"-" ^ s
			else
				s
		in
		Readable_types.IConst s
	| IVar tv ->
		let tv = conv_type_variable tvs tv in
		Readable_types.IVar tv

let conv_arith_op = function
	| A_ADD -> Readable_types.A_ADD
	| A_MUL -> Readable_types.A_MUL

let conv_cmp_op_raw = function
	| C_EQ -> Readable_types.C_EQ
	| C_NEQ -> Readable_types.C_NEQ
	| C_LT -> Readable_types.C_LT
	| C_GT -> Readable_types.C_GT
	| C_LE -> Readable_types.C_LE
	| C_GE -> Readable_types.C_GE

let conv_cmp_op cop =
	let (cop_r, signed) = cop in
	let cop_r = conv_cmp_op_raw cop_r in
	(cop_r, signed)

let rec conv_integer_expr tvs = function
	| IEInte inte ->
		let inte = conv_integer tvs inte in
		Readable_types.IEInte inte
	| IEArith (aop, inte_e1, inte_e2) ->
		let aop = conv_arith_op aop in
		let inte_e1 = conv_integer_expr tvs inte_e1 in
		let inte_e2 = conv_integer_expr tvs inte_e2 in
		Readable_types.IEArith (aop, inte_e1, inte_e2)

let conv_integer_cstrt tvs = function
	| ICCmp (cop, inte_e1, inte_e2) ->
		let cop = conv_cmp_op cop in
		let inte_e1 = conv_integer_expr tvs inte_e1 in
		let inte_e2 = conv_integer_expr tvs inte_e2 in
		Readable_types.ICCmp (cop, inte_e1, inte_e2)

let conv_integer_cstrts tvs ics =
	List.map (conv_integer_cstrt tvs) ics


let conv_kind = function
	| KStack -> Readable_types.KStack
	| KInt -> Readable_types.KInt
	| KStore -> Readable_types.KStore

let conv_kind_to_tv (kind, name) =
	let tv = Readable_types.TV(name) in
	let kind = conv_kind kind in
	(tv, kind)

let conv_kinds_to_tvs kinds =
	List.rev_map conv_kind_to_tv kinds

let rec rename_tv tvs tv n =
	let tv' =
		if n = 0 then
			tv
		else
			let Readable_types.TV(name) = tv in
			Readable_types.TV(name ^ "_" ^ (string_of_int n)) in

	if List.mem tv' tvs then
		rename_tv tvs tv (n + 1)
	else
		tv'

let rec rename_and_extend_tvs tvs c_tvs =
	match c_tvs with
	| [] ->
		(tvs, [])
	| (tv, kind) :: c_tvs ->
		let tv = rename_tv tvs tv 0 in
		let tvs = tv :: tvs in
		let tvs, c_tvs = rename_and_extend_tvs tvs c_tvs in
		tvs, ((tv, kind) :: c_tvs)

let rec conv_stack_type tvs = function
	| StkNull ->
		Readable_types.StkNull
	| StkVar tv ->
		let tv = conv_type_variable tvs tv in
		Readable_types.StkVar tv
	| StkCons (inte, st) ->
		let inte = conv_integer tvs inte in
		let st = conv_stack_type tvs st in
		Readable_types.StkCons (inte, st)

let conv_rich_stack_type tvs rstk =
	let stk = rstk.rstk_stk in
	let freelen = rstk.rstk_freelen in

	let stk = conv_stack_type tvs stk in
	let freelen = conv_integer tvs freelen in

	{ Readable_types.rstk_stk = stk; Readable_types.rstk_freelen = freelen; }

let conv_segment_size_type tvs = function
	| SegSize16 ->
		Readable_types.SegSize16
	| SegSize32 ->
		Readable_types.SegSize32

let conv_segment_type tvs = function
	| CodeSeg sz ->
		let sz = conv_segment_size_type tvs sz in
		Readable_types.CodeSeg sz
	| DataSeg sz ->
		let sz = conv_segment_size_type tvs sz in
		Readable_types.DataSeg sz

let conv_register_kind = function
	| RK_L -> Readable_types.RK_L
	| RK_H -> Readable_types.RK_H
	| RK_X -> Readable_types.RK_X
	| RK_E -> Readable_types.RK_E
	
let conv_register_type tvs reg_type =
	let (reg_kind, inte) = reg_type in
	let reg_kind = conv_register_kind reg_kind in
	let inte = conv_integer tvs inte in
	(reg_kind, inte)

let conv_flags_type tvs flags_type =
{
	Readable_types.f_rst = conv_integer tvs flags_type.f_rst;
	Readable_types.f_op1 = conv_integer tvs flags_type.f_op1;
	Readable_types.f_op2 = conv_integer tvs flags_type.f_op2;
	Readable_types.f_size = flags_type.f_size;
}

let conv_flags_type_opt tvs = function
	| None ->
		None
	| Some flags ->
		let flags = conv_flags_type tvs flags in
		Some flags

let conv_registers_type tvs regs_type =
{
	Readable_types.cs = conv_segment_type tvs regs_type.cs;
(*
	Readable_types.eax = conv_integer tvs reg_type.eax;
	Readable_types.ecx = conv_integer tvs reg_type.ecx;
	Readable_types.edx = conv_integer tvs reg_type.edx;
	Readable_types.ebx = conv_integer tvs reg_type.ebx;
	Readable_types.esi = conv_integer tvs reg_type.esi;
	Readable_types.edi = conv_integer tvs reg_type.edi;
	Readable_types.ebp = conv_integer tvs reg_type.ebp;
	Readable_types.esp = conv_integer tvs reg_type.esp;

	Readable_types.frt = conv_integer tvs reg_type.frt;
	Readable_types.fo1 = conv_integer tvs reg_type.fo1;
	Readable_types.fo2 = conv_integer tvs reg_type.fo2;
*)
	Readable_types.eax = conv_register_type tvs regs_type.eax;
	Readable_types.ecx = conv_register_type tvs regs_type.ecx;
	Readable_types.edx = conv_register_type tvs regs_type.edx;
	Readable_types.ebx = conv_register_type tvs regs_type.ebx;
	Readable_types.esi = conv_register_type tvs regs_type.esi;
	Readable_types.edi = conv_register_type tvs regs_type.edi;
	Readable_types.ebp = conv_register_type tvs regs_type.ebp;
	Readable_types.esp = conv_register_type tvs regs_type.esp;

	Readable_types.flags_opt = conv_flags_type_opt tvs regs_type.flags_opt;
}

let conv_tuple_type_elem tvs (sz, inte) =
	let inte = conv_integer tvs inte in
	(sz, inte)

let conv_tuple_type tvs tt =
	List.map (conv_tuple_type_elem tvs) tt

let rec conv_store tvs str =
	let locs = str.str_locs in
	let vs = str.str_vs in

	let str = [] in
	let str = TVSet.fold (fun tv str ->
		let tv = conv_type_variable tvs tv in
		let se = Readable_types.SEVar tv in
		se :: str
	) vs str in
	let str = IntMap.fold (fun inte at str ->
		let atom = conv_atom tvs (inte, at) in
		let se = Readable_types.SEAtom atom in
		se :: str
	) locs str in

	str

and conv_code_store tvs c_str =
	let str = IntMap.fold (fun inte st str ->
		let inte = conv_integer tvs inte in
		let st = conv_code_type tvs st in
		(inte, st) :: str
	) c_str [] in

	str

and conv_atom tvs (inte, ht) =
	let inte = conv_integer tvs inte in
	let ht = conv_heap_type tvs ht in
	(inte, ht)

and conv_cstrt tvs cstrt =

	let ics = conv_integer_cstrts tvs cstrt.c_ics in
	let store = conv_store tvs cstrt.c_store in
	let code_store = conv_code_store tvs cstrt.c_code_store in

	let cstrt =
	{
		Readable_types.c_ics = ics;
		Readable_types.c_store = store;
		Readable_types.c_code_store = code_store;
	} in
	cstrt

and conv_cstrts tvs cstrts =
	List.map (conv_cstrt tvs) cstrts

and conv_prefix tvs prefix =

	let kinds = prefix.p_kinds in

	let p_tvs = conv_kinds_to_tvs kinds in

	(* solving name conflicts *)
	let tvs, p_tvs = rename_and_extend_tvs tvs p_tvs in

	let cstrts = conv_cstrts tvs prefix.p_cstrts in

	let prefix =
	{
		Readable_types.p_tvs = p_tvs;
		Readable_types.p_cstrts = cstrts;
	} in
	(tvs, prefix)
	
and conv_state_type tvs st =
	let {
		st_prefix = prefix;
		st_registers = reg_typ;
	} = st in

	let (tvs, prefix) = conv_prefix tvs prefix in
	let reg_typ = conv_registers_type tvs reg_typ in

	{ Readable_types.st_prefix = prefix; Readable_types.st_registers = reg_typ; }

and conv_pack_type tvs pt =
	let {
		pt_prefix = prefix;
		pt_tuple = tt;
	} = pt in

	let (tvs, prefix) = conv_prefix tvs prefix in
	let tt = conv_tuple_type tvs tt in

	{ Readable_types.pt_prefix = prefix; Readable_types.pt_tuple = tt; }

and conv_concrete_type tvs = function
	| Pack pt ->
		let pt = conv_pack_type tvs pt in
		Readable_types.Pack pt
	| Tuple tt ->
		let tt = conv_tuple_type tvs tt in
		Readable_types.Tuple tt

and conv_type tvs = function
	| Name nt ->
		let nt = conv_named_type tvs nt in
		Readable_types.Name nt
	| Concrete ct ->
		let ct = conv_concrete_type tvs ct in
		Readable_types.Concrete ct

and conv_array_type tvs at =
	let t = at.at_type in
	let len = at.at_len in

	let t = conv_type tvs t in
	let len = conv_integer tvs len in

	{ Readable_types.at_type = t; Readable_types.at_len = len; }

and conv_heap_type tvs = function
	| Array at ->
		let at = conv_array_type tvs at in
		Readable_types.Array at
	| Stack rstk ->
		let rstk = conv_rich_stack_type tvs rstk in	
		Readable_types.Stack rstk

and conv_code_type tvs = function
	| NameCode nt ->
		let nt = conv_named_type tvs nt in
		Readable_types.NameCode nt
	| State st ->
		let st = conv_state_type tvs st in	
		Readable_types.State st

and conv_named_type tvs nt =
	let (n, cs) = nt in
	let cs = conv_cs tvs cs in
	(n, cs)

and conv_constructor tvs = function
	| CStack st ->
		let st = conv_stack_type tvs st in
		Readable_types.CStack st
	| CInt inte ->
		let inte = conv_integer tvs inte in
		Readable_types.CInt inte
	| CStore str ->
		let str = conv_store tvs str in
		Readable_types.CStore str

and conv_cs tvs cs =
	List.map (conv_constructor tvs) cs

let convert_state state =
	let state = conv_state_type [] state in
	let (tvs, _) = List.split state.Readable_types.st_prefix.Readable_types.p_tvs in
	let tvs = List.rev tvs in
	(state, tvs)

let convert_any converter state a =
	let (state, tvs) = convert_state state in
	let a = converter tvs a in
	(state, a)

let convert_tv =
	convert_any conv_type_variable

let convert_cs =
	convert_any conv_cs

let convert_integer =
	convert_any conv_integer

let convert_integer_expr =
	convert_any conv_integer_expr

let convert_integer_cstrt =
	convert_any conv_integer_cstrt

let convert_integer_cstrts =
	convert_any conv_integer_cstrts

let convert_store =
	convert_any conv_store

let convert_code_store =
	convert_any conv_code_store

let convert_registers_type =
	convert_any conv_registers_type

let convert_concrete_type =
	convert_any conv_concrete_type

let convert_named_type =
	convert_any conv_named_type

let convert_pack_type =
	convert_any conv_pack_type

let convert_type =
	convert_any conv_type

let convert_array_type =
	convert_any conv_array_type

let convert_code_type =
	convert_any conv_code_type
