(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Machine
open Instructions
open Instructions_util

open Disassemble_prefix
open Disassemble_tables
open Disassemble_common

open Util

(*************************************
	Disassemble Modrm
 *************************************)

type modrm =
{
	modrm_mod	: int;
	modrm_rm	: int;
	modrm_reg	: int;
}

let get_modrm b =

	let mod_v = (b lsr 6) land 0x03 in
	let rm_v  = b land 0x07 in
	let reg_v = (b lsr 3) land 0x07 in

	{ modrm_mod = mod_v; modrm_rm = rm_v; modrm_reg = reg_v; }

let get_disp_opt32 md seg off =
	if md = 0 then off, None
	else if md = 1 then
		let disp = get_sint8 seg off in
		let disp_opt = Some disp in
		(off + 1), disp_opt
	else if md = 2 then
		let disp = get_sint32 seg off in
		let disp_opt = Some disp in
		(off + (32 / bits_per_byte)), disp_opt
	else
		failwithf "get_disp_opt32: invalid md = %d > 2" md
		
let get_scaled_index_opt ss index =
	if index = 4 then None
	else
		let reg = get_genreg OpSize32 index in
		let scale = 1 lsl ss in
		Some { msi_index = reg; msi_scale = scale; }

let get_mop_no_sib op_size reg disp_opt =
	Op_Mem {
		mop_base_opt = Some reg;
		mop_disp_opt = disp_opt;
		mop_scaled_index_opt = None;
		mop_size = op_size;
	}

let get_mop_with_sib op_size b disp_opt seg off =
	let sib = get_modrm b in

	let ss = sib.modrm_mod in
	let index = sib.modrm_rm in
	let base = sib.modrm_reg in

	let scaled_index_opt = get_scaled_index_opt ss index in

	if base = 5 && disp_opt = None then
		let disp = get_sint32 seg off in
		(off + 4), get_mem_op op_size None (Some disp) scaled_index_opt
	else
		let r = get_genreg OpSize32 base in
		off, get_mem_op op_size (Some r) disp_opt scaled_index_opt

let get_ea32 op_size modrm seg off =
	let md = modrm.modrm_mod in
	let rm = modrm.modrm_rm in

	if md = 3 then
		off, get_genreg_op op_size rm

	else if md = 0 && rm = 5 then
		let disp = get_sint32 seg off in
		(off + (32 / bits_per_byte)),
		Op_Mem {
			mop_base_opt = None;
			mop_disp_opt = Some disp;
			mop_scaled_index_opt = None;
			mop_size = op_size;
		}

	else if rm = 4 then
		let sib = get_byte seg off in
		let off, disp_opt = get_disp_opt32 md seg (off + 1) in
		get_mop_with_sib op_size sib disp_opt seg off

	else
		let off, disp_opt = get_disp_opt32 md seg off in
		let r = get_genreg OpSize32 rm in
		off, (get_mop_no_sib op_size r disp_opt)

let get_ea dctx op_size modrm seg off =
	match dctx.dctx_ad_attr with
	| AdSizeAttr16 -> failwithf "Address size 16 have not been supported."
	| AdSizeAttr32 -> get_ea32 op_size modrm seg off

let get_operand_E dctx ot modrm seg off =
	get_ea dctx (ot_to_op_size ot dctx.dctx_op_attr) modrm seg off

let get_operand_G dctx ot modrm seg off =
	off, get_genreg_op (ot_to_op_size ot dctx.dctx_op_attr) modrm.modrm_reg

let get_operand_S dctx ot modrm seg off =
	match ot with
	| OT_w -> off, get_segreg_op modrm.modrm_reg
	| _ -> failwithf "Segment register operand doesn't support operand type (%a)." output_optype ot

let get_operand dctx modrm top seg off =
	match top with
	| (AM_E, ot) -> get_operand_E dctx ot modrm seg off
	| (AM_G, ot) -> get_operand_G dctx ot modrm seg off
	| (AM_S, ot) -> get_operand_S dctx ot modrm seg off
	| (a, _) -> failwithf "get_operand_adsize_32: (%a) has not been supported yet." output_admethod a

let get_disas_gen_1op inst top1 =
	fun dctx modrm seg off ->
		let off, op = get_operand dctx modrm top1 seg off in
		off, (inst op)

let get_disas_gen_2op inst top1 top2 =
	fun dctx modrm seg off ->
		let off1, op1 = get_operand dctx modrm top1 seg off in
		let off2, op2 = get_operand dctx modrm top2 seg off in
		(max off1 off2), (inst op1 op2)

let get_disas_ADD top1 top2	= get_disas_gen_2op inst_ADD top1 top2
let get_disas_OR top1 top2	= get_disas_gen_2op inst_OR top1 top2
let get_disas_ADC top1 top2	= get_disas_gen_2op inst_ADC top1 top2
let get_disas_SBB top1 top2	= get_disas_gen_2op inst_SBB top1 top2
let get_disas_AND top1 top2	= get_disas_gen_2op inst_AND top1 top2
let get_disas_SUB top1 top2	= get_disas_gen_2op inst_SUB top1 top2
let get_disas_XOR top1 top2	= get_disas_gen_2op inst_XOR top1 top2
let get_disas_CMP top1 top2	= get_disas_gen_2op inst_CMP top1 top2
let get_disas_BOUND top1 top2	= get_disas_gen_2op inst_BOUND top1 top2
let get_disas_ARPL top1 top2	= get_disas_gen_2op inst_ARPL top1 top2
let get_disas_TEST top1 top2	= get_disas_gen_2op inst_TEST top1 top2
let get_disas_XCHG top1 top2	= get_disas_gen_2op inst_XCHG top1 top2
let get_disas_MOV top1 top2	= get_disas_gen_2op inst_MOV top1 top2
let get_disas_LEA top1 top2	= get_disas_gen_2op inst_LEA top1 top2
let get_disas_LES top1 top2	= get_disas_gen_2op inst_LES top1 top2
let get_disas_LDS top1 top2	= get_disas_gen_2op inst_LDS top1 top2

let get_disas_IMUL top1 top2 top3 =
	fun dctx modrm seg off ->
		let off1, op1 = get_operand dctx modrm top1 seg off in
		let off2, op2 = get_operand dctx modrm top2 seg off in
		let off = max off1 off2 in
		let (off', op3) =
			match top3 with
			| (AM_I, ot) -> get_operand_I dctx ot seg off
			| (a, _) -> failwithf "invalid IMUL operand (%a)" output_admethod a in
		off', (inst_IMUL op1 op2 op3)

let get_disas_POP top		= get_disas_gen_1op inst_POP top
let get_disas_INC top		= get_disas_gen_1op inst_INC top
let get_disas_DEC top		= get_disas_gen_1op inst_DEC top
let get_disas_PUSH top		= get_disas_gen_1op inst_PUSH top
let get_disas_CALLN top		= get_disas_gen_1op inst_CALLN top
let get_disas_CALLF top		= get_disas_gen_1op inst_CALLF top
let get_disas_JMPN top		= get_disas_gen_1op inst_JMPN top
let get_disas_JMPF top		= get_disas_gen_1op inst_JMPF top

let get_disas_Grp1 top1 top2 = fun dctx modrm seg off ->
	let inst = match modrm.modrm_reg with
	| 0 -> inst_ADD
	| 1 -> inst_OR
	| 2 -> inst_ADC
	| 3 -> inst_SBB
	| 4 -> inst_AND
	| 5 -> inst_SUB
	| 6 -> inst_XOR
	| 7 -> inst_CMP
	| _ -> failwithf "invalid Grp1 opcode extension = 0x%02x" modrm.modrm_reg
	in
	let off1, op1 = get_operand dctx modrm top1 seg off in
	let off', op2 =
		match top2 with
		| (AM_I, ot) -> get_operand_I dctx ot seg off1
		| (a, _) -> failwithf "invalid Grp1 operand (%a)." output_admethod a in
	off', (inst op1 op2)

let get_disas_Grp5 = fun dctx modrm seg off ->
	let f = match modrm.modrm_reg with
	| 0 -> get_disas_INC oi_Ev
	| 1 -> get_disas_DEC oi_Ev
	| 2 -> get_disas_CALLN oi_Ev
	| 3 -> get_disas_CALLF oi_Ep
	| 4 -> get_disas_JMPN oi_Ev
	| 5 -> get_disas_JMPF oi_Ep
	| 6 -> get_disas_PUSH oi_Ev
	| _ -> failwithf "invalid Grp5 opcode extension = 0x%02x" modrm.modrm_reg
	in
	f dctx modrm seg off

let get_disas_Grp11 top1 top2 =
	fun dctx modrm seg off ->
		let off1, op1 = get_operand dctx modrm top1 seg off in
		let off', op2 =
			match top2 with
			| (AM_I, ot) -> get_operand_I dctx ot seg off1
			| (a, _) -> failwithf "invalid Grp11 operand (%a)." output_admethod a in
		off', (inst_MOV op1 op2)

let get_1byte_modrm_op_raw b =
	match b with
	| 0x00 -> get_disas_ADD		oi_Eb oi_Gb
	| 0x01 -> get_disas_ADD		oi_Ev oi_Gv
	| 0x02 -> get_disas_ADD		oi_Gb oi_Eb
	| 0x03 -> get_disas_ADD		oi_Gv oi_Ev
	| 0x08 -> get_disas_OR		oi_Eb oi_Gb
	| 0x09 -> get_disas_OR		oi_Ev oi_Gv
	| 0x0A -> get_disas_OR		oi_Gb oi_Eb
	| 0x0B -> get_disas_OR		oi_Gv oi_Ev
	| 0x10 -> get_disas_ADC		oi_Eb oi_Gb
	| 0x11 -> get_disas_ADC		oi_Ev oi_Gv
	| 0x12 -> get_disas_ADC		oi_Gb oi_Eb
	| 0x13 -> get_disas_ADC		oi_Gv oi_Ev
	| 0x18 -> get_disas_SBB		oi_Eb oi_Gb
	| 0x19 -> get_disas_SBB		oi_Ev oi_Gv
	| 0x1A -> get_disas_SBB		oi_Gb oi_Eb
	| 0x1B -> get_disas_SBB		oi_Gv oi_Ev
	| 0x20 -> get_disas_AND		oi_Eb oi_Gb
	| 0x21 -> get_disas_AND		oi_Ev oi_Gv
	| 0x22 -> get_disas_AND		oi_Gb oi_Eb
	| 0x23 -> get_disas_AND		oi_Gv oi_Ev
	| 0x28 -> get_disas_SUB		oi_Eb oi_Gb
	| 0x29 -> get_disas_SUB		oi_Ev oi_Gv
	| 0x2A -> get_disas_SUB		oi_Gb oi_Eb
	| 0x2B -> get_disas_SUB		oi_Gv oi_Ev
	| 0x30 -> get_disas_XOR		oi_Eb oi_Gb
	| 0x31 -> get_disas_XOR		oi_Ev oi_Gv
	| 0x32 -> get_disas_XOR		oi_Gb oi_Eb
	| 0x33 -> get_disas_XOR		oi_Gv oi_Ev
	| 0x38 -> get_disas_CMP		oi_Eb oi_Gb
	| 0x39 -> get_disas_CMP		oi_Ev oi_Gv
	| 0x3A -> get_disas_CMP		oi_Gb oi_Eb
	| 0x3B -> get_disas_CMP		oi_Gv oi_Ev
	| 0x62 -> get_disas_BOUND	oi_Gv oi_Ma
	| 0x63 -> get_disas_ARPL	oi_Ew oi_Gw
	| 0x69 -> get_disas_IMUL	oi_Gv oi_Ev oi_Iv
	| 0x6B -> get_disas_IMUL	oi_Gv oi_Ev oi_Ib
	| 0x80 -> get_disas_Grp1	oi_Eb oi_Ib
	| 0x81 -> get_disas_Grp1	oi_Ev oi_Iv
	| 0x82 -> get_disas_Grp1	oi_Eb oi_Ib	(* alias for 0x80 ... *)
	| 0x83 -> get_disas_Grp1	oi_Ev oi_Ib
	| 0x84 -> get_disas_TEST	oi_Eb oi_Gb
	| 0x85 -> get_disas_TEST	oi_Ev oi_Gv
	| 0x86 -> get_disas_XCHG	oi_Eb oi_Gb
	| 0x87 -> get_disas_XCHG	oi_Ev oi_Gv
	| 0x88 -> get_disas_MOV		oi_Eb oi_Gb
	| 0x89 -> get_disas_MOV		oi_Ev oi_Gv
	| 0x8A -> get_disas_MOV		oi_Gb oi_Eb
	| 0x8B -> get_disas_MOV		oi_Gv oi_Ev
	| 0x8C -> get_disas_MOV		oi_Ew oi_Sw
	| 0x8D -> get_disas_LEA		oi_Gv oi_M
	| 0x8E -> get_disas_MOV		oi_Sw oi_Ew
	| 0x8F -> get_disas_POP		oi_Ev
(*
	| 0xC0 -> get_disas_Grp2	oi_Eb oi_Ib
	| 0xC1 -> get_disas_Grp2	oi_Ev oi_Ib
*)
	| 0xC4 -> get_disas_LES		oi_Gv oi_Mp
	| 0xC5 -> get_disas_LDS		oi_Gv oi_Mp
	| 0xC6 -> get_disas_Grp11	oi_Eb oi_Ib
	| 0xC7 -> get_disas_Grp11	oi_Ev oi_Iv
(*
	| 0xD0 -> get_disas_Grp2	oi_Eb oi_1
	| 0xD1 -> get_disas_Grp2	oi_Ev oi_1
	| 0xD2 -> get_disas_Grp2	oi_Eb oi_CL
	| 0xD3 -> get_disas_Grp2	oi_Ev oi_CL
	| 0xF6 -> get_disas_Grp3	oi_Eb
	| 0xF7 -> get_disas_Grp3	oi_Ev
	| 0xFE -> get_disas_Grp4
*)
	| 0xFF -> get_disas_Grp5
	| _ -> reserved_inst "Unknown opcode = 0x%02x" b

let get_2byte_modrm_op_0F_raw b =
	match b with
	| _ -> reserved_inst "Unknown opcode = 0x0f%02x" b

let get_1byte_modrm_op b =
	let disas_raw = get_1byte_modrm_op_raw b in
	fun dctx seg off ->
		let b = get_byte seg off in
		let modrm = get_modrm b in
		let off = off + 1 in
		disas_raw dctx modrm seg off

let get_2byte_modrm_op_0F b =
	let disas_raw = get_2byte_modrm_op_0F_raw b in
	fun dctx seg off ->
		let b = get_byte seg off in
		let modrm = get_modrm b in
		let off = off + 1 in
		disas_raw dctx modrm seg off
