(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Nativeint
open Machine
open Elf_common
open Elf_header

open Util

type section_header =
{
	sh_name		: elf32_word;
	sh_type		: section_type;
	sh_flags	: section_flags;
	sh_addr		: elf32_addr;
	sh_offset	: elf32_off;
	sh_size		: elf32_word;
	sh_link		: elf32_word;
	sh_info		: elf32_word;
	sh_addralign	: elf32_word;
	sh_entsize	: elf32_word;
}

and section_type =
	| SHT_PROGBITS
	| SHT_SYMTAB
	| SHT_STRTAB
	| SHT_RELA
	| SHT_HASH
	| SHT_DYNAMIC
	| SHT_NOTE
	| SHT_NOBITS
	| SHT_REL
	| SHT_SHLIB
	| SHT_DYNSYM
	| SHT_VERSYM
	| SHT_VERNEED

and section_flags =
{
	shf_write	: bool;
	shf_alloc	: bool;
	shf_execinstr	: bool;
}

let fi_name		= { fi_off = 0;				fi_size = elf32_word_size; }
let fi_type		= { fi_off = next_off fi_name;		fi_size = elf32_word_size; }
let fi_flags		= { fi_off = next_off fi_type;		fi_size = elf32_word_size; }
let fi_addr		= { fi_off = next_off fi_flags;		fi_size = elf32_addr_size; }
let fi_offset		= { fi_off = next_off fi_addr;		fi_size = elf32_off_size; }
let fi_size		= { fi_off = next_off fi_offset;	fi_size = elf32_word_size; }
let fi_link		= { fi_off = next_off fi_size;		fi_size = elf32_word_size; }
let fi_info		= { fi_off = next_off fi_link;		fi_size = elf32_word_size; }
let fi_addralign	= { fi_off = next_off fi_info;		fi_size = elf32_word_size; }
let fi_entsize		= { fi_off = next_off fi_addralign;	fi_size = elf32_word_size; }

let shf_write_bit = 0x1
let shf_alloc_bit = 0x2
let shf_execinstr_bit = 0x4

let get_name seg = get_field seg fi_name

let get_type seg =
	let v = get_field seg fi_type in
	let i = to_int v in
	match i with
	| 1 -> SHT_PROGBITS
	| 2 -> SHT_SYMTAB
	| 3 -> SHT_STRTAB
	| 4 -> SHT_RELA
	| 5 -> SHT_HASH
	| 6 -> SHT_DYNAMIC
	| 7 -> SHT_NOTE
	| 8 -> SHT_NOBITS
	| 9 -> SHT_REL
	| 10 -> SHT_SHLIB
	| 11 -> SHT_DYNSYM
	| 0x6ffffffe -> SHT_VERNEED
	| 0x6fffffff -> SHT_VERSYM
	| _ -> failwithf "invalid ELF section type = %d." i

let get_flags seg =
	let v = get_field seg fi_flags in
	let i = to_int v in
	let s_write = (i land shf_write_bit) <> 0 in
	let s_alloc = (i land shf_alloc_bit) <> 0 in
	let s_execinstr = (i land shf_execinstr_bit) <> 0 in
	{
		shf_write = s_write;
		shf_alloc = s_alloc;
		shf_execinstr = s_execinstr;
	}

let get_addr seg = get_field seg fi_addr
let get_offset seg = get_field seg fi_offset
let get_size seg = get_field seg fi_size
let get_link seg = get_field seg fi_link
let get_info seg = get_field seg fi_info
let get_addralign seg = get_field seg fi_addralign
let get_entsize seg = get_field seg fi_entsize

let get_section_header seg =
	let sh_name = get_name seg in
	let sh_type = get_type seg in
	let sh_flags = get_flags seg in
	let sh_addr = get_addr seg in
	let sh_offset = get_offset seg in
	let sh_size = get_size seg in
	let sh_link = get_link seg in
	let sh_info = get_info seg in
	let sh_addralign = get_addralign seg in
	let sh_entsize = get_entsize seg in
	{
		sh_name = sh_name;
		sh_type = sh_type;
		sh_flags = sh_flags;
		sh_addr = sh_addr;
		sh_offset = sh_offset;
		sh_size = sh_size;
		sh_link = sh_link;
		sh_info = sh_info;
		sh_addralign = sh_addralign;
		sh_entsize = sh_entsize;
	}

let get_section_headers header seg =
	let off = to_int header.e_shoff in
	let entsize = to_int header.e_shentsize in
	let num = to_int header.e_shnum in

	(* Skip the first section headers because it is NULL section. *)
	let off = off + entsize in
	let num = num - 1 in

	get_headers_common seg off entsize num get_section_header

open Buffer

let rec get_null_terminated_string_buffer seg off buf =
	let c = Char.chr (get_byte seg off) in
	if c = '\000' then ()
	else
		(add_char buf c;
		get_null_terminated_string_buffer seg (off + 1) buf)

let get_null_terminated_string seg off =
	let buf = Buffer.create 8 in
	get_null_terminated_string_buffer seg off buf;
	contents buf

let get_section_name section strtab_seg =
	let index = to_int section.sh_name in

	get_null_terminated_string strtab_seg index

let get_string_table header section_headers seg =
	let shstrndx = to_int header.e_shstrndx in
	let section_header = List.nth section_headers (shstrndx - 1) in
	if section_header.sh_type <> SHT_STRTAB then failwithf "invalid ELF header (String Table Index is invalid).";
	{
		seg_va = seg.seg_va;
		seg_array = seg.seg_array;
		seg_offset = to_int section_header.sh_offset;
		seg_size = to_int section_header.sh_size;
	}

let get_section_seg section_header seg =
	let off = to_int section_header.sh_offset in
	let size = to_int section_header.sh_size in

	get_subseg seg off size
