(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Types

open Util

open Format

open Big_int
open Big_int_util

(**********************************************
	Routine for Printing Types (for DEBUG)
 **********************************************)

let rec output_list ppf lst delim output_elm =
	let rec aux first lst =
		match lst with
		| [] -> ()
		| elm :: lst ->
		(if first then
			fprintf ppf "%a" output_elm elm
		else
			fprintf ppf "%s@ %a" delim output_elm elm);
		aux false lst
	in
	aux true lst

let rec output_kind ppf k =
	match k with
	| KStack -> fprintf ppf "STACK"
	| KInt -> fprintf ppf "INT"
	| KStore -> fprintf ppf "STORE"

and output_type ppf t =
	match t with
	| Name nt ->
		fprintf ppf "%a" output_nt nt
	| Concrete (ct) ->
		fprintf ppf "%a" output_ct ct

and output_at ppf at =
	let t = at.at_type in
	let len = at.at_len in

	fprintf ppf "%a" output_type t;

	match len with
	| IConst n when eq_big_int n unit_big_int ->
		()
	| _ ->
		fprintf ppf "[%a]" output_inte len

and output_code ppf ct =
	match ct with
	| NameCode nt ->
		fprintf ppf "%a" output_nt nt
	| State st ->
		fprintf ppf "%a" output_state st

and output_nt ppf nt =
	let (n, cs) = nt in
	fprintf ppf "%s(" n;
	fprintf ppf "%a" output_cs cs;
	fprintf ppf ")"

and output_ct ppf ct =
	match ct with
	(* De Bruijn's notation! *)
	| Pack pt ->
		fprintf ppf "%a" output_pt pt
	| Tuple tt ->
		fprintf ppf "%a" output_tt tt

and output_tt ppf tt =
	fprintf ppf "@[<3><%a>@]" output_tes tt

and output_tes ppf tes =
	output_list ppf tes "," output_te

and output_te ppf (sz, inte) =
	match sz with
	| Instructions.OpSize8 ->
		fprintf ppf "B8(%a)" output_inte inte
	| Instructions.OpSize16 ->
		fprintf ppf "B16(%a)" output_inte inte
	| Instructions.OpSize32 ->
		fprintf ppf "%a" output_inte inte
	| Instructions.OpSize48 ->
		fprintf ppf "B48(%a)" output_inte inte
	| Instructions.OpSize64 ->
		fprintf ppf "B64(%a)" output_inte inte
	| Instructions.OpSize128 ->
		fprintf ppf "B128(%a)" output_inte inte

and output_pt ppf pt =
	let { pt_prefix = prefix; pt_tuple = tt; } = pt in
	fprintf ppf "@[<1>Exists@,%a@ %a@]" output_prefix prefix output_tt tt

and output_state ppf st =
	let { st_prefix = prefix; st_registers = regs_type; } = st in
	fprintf ppf "@[<1>Forall@,%a@ %a@]" output_prefix prefix output_regs_type regs_type

and output_prefix ppf prefix =
	let kinds = prefix.p_kinds in
	let cstrts = prefix.p_cstrts in

	fprintf ppf "@[<1>";
	List.iter (fun (kind, name) -> fprintf ppf "(%s):%a;@ " name output_kind kind) kinds;
	fprintf ppf "%a" output_cstrts cstrts;
	fprintf ppf "@]"

and output_cstrts ppf cstrts =
	List.iter (fun cstrt -> fprintf ppf "%a@ " output_cstrt cstrt) cstrts

and output_cstrt ppf cstrt =
	let ics = cstrt.c_ics in
	let store = cstrt.c_store in
	let code_store = cstrt.c_code_store in

	let ics_empty = ics <> [] in
	let str_empty = IntMap.is_empty store.str_locs && TVSet.is_empty store.str_vs in
	let cstr_empty = IntMap.is_empty code_store in

	fprintf ppf "@[<1>[";
	fprintf ppf "%a" output_inte_cstrts ics;
	if ics_empty && not (str_empty && cstr_empty) then fprintf ppf "|@ ";
	fprintf ppf "%a" output_store store;
	if str_empty && not cstr_empty then fprintf ppf "|@ ";
	fprintf ppf "%a" output_code_store code_store;
	fprintf ppf "]@]"

and output_cs ppf cs =
	List.iter (fprintf ppf "%a " output_cstr) cs

and output_regs_type ppf regs_type =
	fprintf ppf "@[<1>{@,";
(*
	fprintf ppf "eax = %a;@ " output_inte reg_type.eax;
	fprintf ppf "ecx = %a;@ " output_inte reg_type.ecx;
	fprintf ppf "edx = %a;@ " output_inte reg_type.edx;
	fprintf ppf "ebx = %a;@ " output_inte reg_type.ebx;
	fprintf ppf "esi = %a;@ " output_inte reg_type.esi;
	fprintf ppf "edi = %a;@ " output_inte reg_type.edi;
	fprintf ppf "ebp = %a;@ " output_inte reg_type.ebp;
	fprintf ppf "esp = %a;@ " output_inte reg_type.esp;
*)
	fprintf ppf "eax = %a;@ " output_reg_type regs_type.eax;
	fprintf ppf "ecx = %a;@ " output_reg_type regs_type.ecx;
	fprintf ppf "edx = %a;@ " output_reg_type regs_type.edx;
	fprintf ppf "ebx = %a;@ " output_reg_type regs_type.ebx;
	fprintf ppf "esi = %a;@ " output_reg_type regs_type.esi;
	fprintf ppf "edi = %a;@ " output_reg_type regs_type.edi;
	fprintf ppf "ebp = %a;@ " output_reg_type regs_type.ebp;
	fprintf ppf "esp = %a;@ " output_reg_type regs_type.esp;
	fprintf ppf "}@]";
	fprintf ppf "%a" output_flags_type_opt regs_type.flags_opt

and output_flags_type_opt ppf = function
	| None ->
		()
	| Some flags ->
		output_flags_type ppf flags

and output_flags_type ppf flags_type =
	fprintf ppf "@[<2>{@,";
	fprintf ppf "f_rst = %a;@ " output_inte flags_type.f_rst;
	fprintf ppf "f_op1 = %a;@ " output_inte flags_type.f_op1;
	fprintf ppf "f_op2 = %a;@ " output_inte flags_type.f_op2;
	(* XXX : do not print operand_size *)
	fprintf ppf "}@]"

and output_reg_type ppf reg_type =
	let (reg_kind, inte) = reg_type in
	let suffix = match reg_kind with
		| RK_L -> "(L)"
		| RK_H -> "(H)"
		| RK_X -> "(X)"
		| RK_E -> ""
	in
	fprintf ppf "%a%s" output_inte inte suffix

(*
and output_wts ppf wts =
	output_list ppf wts "," output_inte
*)

and output_stk ppf = function
	| StkNull ->
		fprintf ppf "$"
	| StkVar tv ->
		fprintf ppf "%a" output_tv tv
	| StkCons (inte, stk')->
		fprintf ppf "%a@ ::@ %a" output_inte inte output_stk stk'

and output_rstk ppf rstk =
	let stk = rstk.rstk_stk in
	let freelen = rstk.rstk_freelen in
	fprintf ppf "%a(%a)" output_stk stk output_inte freelen

and output_tv ppf tv =
	let TV(i) = tv in
	fprintf ppf "(X_%d)" i
		
and output_inte ppf = function
	| IConst (n) ->
(*
		fprintf ppf "0x%nx" ni
*)
(*
		fprintf ppf "0x%a" output_big_int_x n
*)
		fprintf ppf "%a" output_big_int_x n
	| IVar (tv) ->
		fprintf ppf "%a" output_tv tv

and output_aop ppf = function
	| A_ADD ->
		fprintf ppf "+"
	| A_MUL ->
		fprintf ppf "*"

and output_cop_raw ppf = function
	| C_EQ ->
		fprintf ppf "="
	| C_NEQ ->
		fprintf ppf "<>"
	| C_LT ->
		fprintf ppf "<"
	| C_GT ->
		fprintf ppf ">"
	| C_LE ->
		fprintf ppf "<="
	| C_GE ->
		fprintf ppf ">="

and output_cop ppf cop =
	let (cop_r, signed) = cop in
	fprintf ppf "%a%s" output_cop_raw cop_r (if signed then "" else "u")

and output_inte_expr ppf = function
	| IEInte inte ->
		fprintf ppf "%a" output_inte inte
	| IEArith (aop, inte_e1, inte_e2) ->
		fprintf ppf "(%a@ %a@ %a)@ " output_inte_expr inte_e1 output_aop aop output_inte_expr inte_e2

and output_inte_cstrt ppf = function
	| ICCmp (cop, inte_e1, inte_e2) ->
		fprintf ppf "@[<5>{%a@ %a@ %a}@;@,@]"
			output_inte_expr inte_e1
			output_cop cop
			output_inte_expr inte_e2

and output_inte_cstrts ppf ics =
	List.iter (output_inte_cstrt ppf) ics

and output_store ppf str =
	let locs = str.str_locs in
	let vs = str.str_vs in

	fprintf ppf "%a%a" output_locs locs output_vs vs

and output_locs ppf locs =
	IntMap.iter (fun inte t ->
		fprintf ppf "%a@ " output_atom (inte, t)
	) locs

and output_vs ppf vs =
	TVSet.iter (fun tv ->
		fprintf ppf "{%a}@ " output_tv tv
	) vs

and output_atom ppf (inte, ht) =
	fprintf ppf "@[<5>{%a@ -->@ %a}@]" output_inte inte output_ht ht

and output_ht ppf = function
	| Array at ->
		output_at ppf at
	| Stack rstk ->
		fprintf ppf "%a" output_rstk rstk

and output_code_store ppf c_str =
	let output_aux ppf e =
		let (inte, st) = e in
		fprintf ppf "@[<5>{%a@ -->@ %a}@]" output_inte inte output_code st
	in

	IntMap.iter (fun inte ct ->
		output_aux ppf (inte, ct)
	) c_str

and output_cstr ppf a =
	match a with
	| CStack st -> fprintf ppf "%a:stack" output_stk st
	| CInt inte -> fprintf ppf "%a:int" output_inte inte
	| CStore str -> fprintf ppf "%a:store" output_store str
