(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Nativeint

module TnativeInt =
struct
	type t = nativeint
	let compare = Nativeint.compare
end

module NativeintMap = Map.Make (TnativeInt)

module Tstring =
struct
	type t = string
	let compare = String.compare
end

module StringMap = Map.Make (Tstring)

open Format

let failwithf_exn c =
	kfprintf
	(fun _ ->
		let str = flush_str_formatter () in
		c str) str_formatter

let failwithf f = failwithf_exn failwith f



(* File to Bigarray.Array1 util *)

let file_to_bigarray fd =
	Bigarray.Array1.map_file fd Bigarray.int8_unsigned Bigarray.c_layout false (-1)

let read_file filename =
	let fd = Unix.openfile filename [ Unix.O_RDONLY ] 0 in
	file_to_bigarray fd



(* List utility functions *)

let rec fold_lefti s f r lst =
	let (_, r) = List.fold_left (fun (i, r) elm -> (i + 1, f i r elm) ) (s, r) lst in
	r

let rec fold_righti s f lst r =
	let (_, r) = List.fold_right (fun elm (i, r) -> (i + 1, f i elm r) ) lst (s, r) in
	r

let rev_mapi s f lst =
	fold_lefti s (fun i r elm -> (f i elm) :: r) [] lst

let mapi s f lst =
	let r = rev_mapi s f lst in
	List.rev r


let iteri s f lst =
	fold_lefti s (fun i () elm -> f i elm) () lst

let rec repeat n f v =
	if n = 0 then v
	else
	let v = f v in
	repeat (n - 1) f v


let rec set_nth lst i e =
	match lst with
	| [] ->
		failwith "nth"
	| e' :: lst ->
		if i = 0 then
			e :: lst
		else
			e' :: (set_nth lst (i - 1) e)

let rec remove_last = function
	| [] ->
		failwith "remove_last"
	| [_] ->
		[]
	| x :: xs ->
		x :: (remove_last xs)

let rec is_equal_unordered is_equal lst1 lst2 =
	match lst1, lst2 with
	| [], [] ->
		true
	| [], _ ->
		false
	| e1 :: lst1, _ ->
		let e2s, lst2 = List.partition (is_equal e1) lst2 in
		begin
			match e2s with
			| [_] ->
				is_equal_unordered is_equal lst1 lst2
			| _ ->
				false
		end

let pop_with_def def = function
	| [] ->
		def, []
	| x :: xs ->
		x, xs

let rec find_index_no_error i elm = function
	| [] ->
		-1
	| elm' :: lst ->
		if elm = elm' then
			i
		else
			find_index_no_error (i + 1) elm lst

let find_index elm lst =
	let i = find_index_no_error 0 elm lst in
	if i < 0 then
		raise Not_found
	else
		i



(* Global counter *)

let c = ref 0

let init_counter v =
	c := v

let get_counter_value () =
	let v = !c in
	c := v + 1;
	v

(* Misc *)
let chop_sign str =
	let len = String.length str in

	if len <= 0 then
		false, str
	else

	let c = str.[0] in

	if c = '-' then
		let str = String.sub str 1 (len - 1) in
		true, str
	else
		false, str

let get_sign str =
	let len = String.length str in

	if len <= 0 then
		false

	else

	let c = str.[0] in

	c = '-'
